"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = void 0;
const RepoManager_1 = require("../manager/RepoManager");
const BelongsTo_1 = require("../relationships/BelongsTo");
const hasOne_1 = require("../relationships/hasOne");
const HasMany_1 = require("../relationships/HasMany");
const BelongsToMany_1 = require("../relationships/BelongsToMany");
const QueryBuilder_1 = require("./QueryBuilder");
const moment_1 = __importDefault(require("moment"));
const pluralize_1 = __importDefault(require("pluralize"));
class Model {
    constructor(attributes) {
        if (attributes)
            this.fill(attributes);
    }
    static get modelName() {
        return (new this).constructor.name;
    }
    get cName() {
        return this.constructor.collectionName || pluralize_1.default(this.constructor.name, 2);
    }
    static get cName() {
        return this.collectionName || pluralize_1.default((new this()).constructor.name, 2);
    }
    get needTimestamp() {
        let timestamp = this.constructor.timestamp;
        if (timestamp === undefined) {
            timestamp = true;
        }
        return timestamp;
    }
    // start of object construction
    fill(attributes) {
        console.log('attributes: ', attributes);
        for (const field in attributes) {
            this[field] = attributes[field];
        }
    }
    replicate() {
        const replicatedModel = Object.assign(Object.create(Object.getPrototypeOf(this)), this);
        delete replicatedModel._id;
        return replicatedModel;
    }
    // end of object construction
    // start of CRUD operation
    static first() {
        // @ts-ignore
        return this.query().first();
    }
    static count() {
        // @ts-ignore
        return this.query().count();
    }
    static all() {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            const items = yield RepoManager_1.RepoManager.get(collectionName).all();
            const result = [];
            for (const item of items) {
                const castedItem = new this();
                castedItem.fill(item);
                result.push(castedItem);
            }
            return result;
        });
    }
    static find(primaryKey) {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            const item = yield RepoManager_1.RepoManager.get(collectionName).find(primaryKey);
            if (!item)
                return undefined;
            const model = new this(item);
            model.fill(item);
            return model;
        });
    }
    static create(attributes) {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            const model = new this(attributes);
            // @ts-ignore
            if (model.needTimestamp) {
                attributes.createdAt = moment_1.default().toISOString();
                attributes.updatedAt = moment_1.default().toISOString();
            }
            const result = yield RepoManager_1.RepoManager.get(collectionName).create(attributes);
            model._id = result.id;
            return model;
        });
    }
    touch() {
        return __awaiter(this, void 0, void 0, function* () {
            const id = this._id;
            const collectionName = this.cName;
            const updatedModel = yield RepoManager_1.RepoManager.get(collectionName).find(id);
            this.fill(updatedModel);
        });
    }
    update(attributes) {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            const guarded = this.constructor.readonlyFields;
            attributes._id = this._id;
            if (this.needTimestamp)
                attributes.updatedAt = moment_1.default().toISOString();
            yield RepoManager_1.RepoManager.get(collectionName).update(attributes, guarded);
            yield this.touch();
            return this;
        });
    }
    save() {
        return __awaiter(this, void 0, void 0, function* () {
            let updatingField = {};
            // const classHasAttribute = (key: string) => Object.keys(new (this.constructor as typeof Model)()).includes(key)
            for (const field in this) {
                if (typeof field === 'function')
                    continue;
                // if (!classHasAttribute(field)) {
                //     throw new Error(`Class "${(this.constructor as typeof Model).name}" does not have attribute "${field}" `);
                // }
                updatingField[field] = this[field];
            }
            const now = moment_1.default().toISOString();
            if (this.needTimestamp)
                updatingField.updatedAt = now;
            const collectionName = this.cName;
            if (!this._id) {
                if (this.needTimestamp)
                    updatingField.createdAt = now;
                const result = yield RepoManager_1.RepoManager.get(collectionName).create(updatingField);
                this._id = result.id;
            }
            else {
                const guarded = this.constructor.readonlyFields;
                yield RepoManager_1.RepoManager.get(collectionName).update(updatingField, guarded);
            }
            yield this.touch();
            return this;
        });
    }
    delete() {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            yield RepoManager_1.RepoManager.get(collectionName).delete(this._id);
            Object.keys(this).forEach((key) => delete this[key]);
        });
    }
    // end of CRUD operation
    // start of query builder
    static query() {
        const collectionName = this.cName;
        console.log('collectionName: ', collectionName);
        const model = new this;
        return QueryBuilder_1.QueryBuilder.query(collectionName, model);
    }
    static where(...args) {
        let field, operator, value;
        if (args.length === 1) {
            // @ts-ignore
            const builder = this.query();
            return builder.whereCondition(args[0]);
        }
        if (args.length === 3) {
            field = args[0];
            operator = args[1];
            value = args[2];
        }
        else {
            field = args[0];
            operator = '=';
            value = args[1];
        }
        // @ts-ignore
        const builder = this.query();
        return builder.where(field, operator, value);
    }
    // end of query builder
    // start of relationship
    static with(...relationships) {
        const collectionName = this.cName;
        const model = new this;
        return QueryBuilder_1.QueryBuilder.query(collectionName, model, relationships);
    }
    load(...relationships) {
        return __awaiter(this, void 0, void 0, function* () {
            const collectionName = this.cName;
            const builder = new QueryBuilder_1.QueryBuilder(collectionName, this, relationships);
            const loadedModel = yield builder.first();
            for (const relationship of relationships) {
                this[relationship] = loadedModel[relationship];
            }
            return this;
        });
    }
    belongsTo(relationship, localKey, foreignKey) {
        return BelongsTo_1.belongsTo(this, relationship, localKey, foreignKey);
    }
    hasOne(relationship, localKey, foreignKey) {
        return hasOne_1.hasOne(this, relationship, localKey, foreignKey);
    }
    hasMany(relationship, localKey, foreignKey) {
        return HasMany_1.hasMany(this, relationship, localKey, foreignKey);
    }
    belongsToMany(relationship, pivot, localKey, foreignKey) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield BelongsToMany_1.belongsToMany(this, relationship, pivot, localKey, foreignKey);
        });
    }
    toResponse() {
        const replicatedModel = Object.assign(Object.create(Object.getPrototypeOf(this)), this);
        delete replicatedModel.save;
        for (const key in replicatedModel) {
            let formattedResult;
            if (Array.isArray(replicatedModel[key]) && replicatedModel[key][0] instanceof Model) {
                // @ts-ignore
                formattedResult = replicatedModel[key].map(m => m.toResponse());
            }
            else if (replicatedModel[key] instanceof Model) {
                // @ts-ignore
                formattedResult = replicatedModel[key].toResponse();
            }
            if (!formattedResult)
                continue;
            replicatedModel[key] = formattedResult;
        }
        if (this.formatResponse) {
            return this.formatResponse(replicatedModel);
        }
        if (!replicatedModel)
            throw new Error(`${this.constructor.name}.formatResponse() must return an object`);
        return replicatedModel;
    }
}
exports.Model = Model;
Model.timestamp = true;
//# sourceMappingURL=Model.js.map