"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QueryBuilder = void 0;
const mongodb_1 = require("mongodb");
const RepoManager_1 = require("../manager/RepoManager");
class QueryBuilder {
    constructor(collectionName, modelClass, relationships, isOne) {
        this.queries = {
            $and: [],
        };
        this.sorter = {};
        if (collectionName === undefined) {
            throw new Error(`QueryBuilder create error: collectionName not found`);
        }
        this.repo = RepoManager_1.RepoManager.get(collectionName);
        this.modelClass = modelClass;
        this.relationships = relationships;
        this.isOne = isOne;
    }
    static query(collectionName, modelClass, relationships) {
        return new this(collectionName, modelClass, relationships);
    }
    static where(field, operator, value, collectionName, modelClass) {
        const builder = this.query(collectionName, modelClass);
        return builder.where(field, operator, value);
    }
    raw() {
        return this.repo.raw();
    }
    where(...args) {
        let field, operator, value;
        if (args.length === 3) {
            field = args[0];
            operator = args[1];
            value = args[2];
        }
        else if (args.length === 2) {
            field = args[0];
            operator = '=';
            value = args[1];
        }
        else {
            return this.whereCondition(args[0]);
        }
        if (mongodb_1.ObjectID.isValid(value) && field.toLocaleLowerCase().includes('id')) {
            value = new mongodb_1.ObjectID(value);
        }
        switch (operator) {
            case '=':
                this.queries.$and.push({ [field]: { $eq: value } });
                break;
            case '!=':
                this.queries.$and.push({ [field]: { $ne: value } });
                break;
            case '>':
                this.queries.$and.push({ [field]: { $gt: value } });
                break;
            case '>=':
                this.queries.$and.push({ [field]: { $gte: value } });
                break;
            case '<':
                this.queries.$and.push({ [field]: { $lt: value } });
                break;
            case '<=':
                this.queries.$and.push({ [field]: { $lte: value } });
                break;
            case 'in':
                this.queries.$and.push({ [field]: { $in: value } });
                break;
            case 'between':
                this.queries.$and.push({ [field]: { $gte: value[0], $lte: value[1] } });
                break;
        }
        this.lastWhere = field;
        return this;
    }
    orWhere(...args) {
        let field, operator, value;
        if (args.length === 3) {
            field = args[0];
            operator = args[1];
            value = args[2];
        }
        else {
            field = args[0];
            operator = '=';
            value = args[1];
        }
        const { lastWhere } = this;
        const queries = this.queries.$and;
        const lastQueryIndex = queries.length - 1;
        const lastQuery = queries[lastQueryIndex];
        this.queries.$and = this.queries.$and.filter((query, i) => i !== lastQueryIndex);
        let newQuery = {};
        switch (operator) {
            case '=':
                newQuery = { [field]: { $eq: value } };
                break;
            case '!=':
                newQuery = { [field]: { $ne: value } };
                break;
            case '>':
                newQuery = { [field]: { $gt: value } };
                break;
            case '>=':
                newQuery = { [field]: { $gte: value } };
                break;
            case '<':
                newQuery = { [field]: { $lt: value } };
                break;
            case '<=':
                newQuery = { [field]: { $lte: value } };
                break;
            case 'in':
                newQuery = { [field]: { $in: value } };
                break;
            case 'between':
                newQuery = { [field]: { $gte: value[0], $lte: value[1] } };
                break;
        }
        if (lastWhere === '$or') {
            lastQuery.$or.push(newQuery);
            this.queries.$and.push(lastQuery);
        }
        else {
            this.queries.$and.push({ $or: [
                    lastQuery,
                    newQuery,
                ] });
        }
        this.lastWhere = '$or';
        return this;
    }
    whereCondition(condition) {
        if (typeof condition === 'function') {
            // @ts-ignore
            const newQueryBuilder = new QueryBuilder(this.modelClass.cName, this.modelClass, []);
            condition(newQueryBuilder);
            console.log('newQueryBuilder.queries.$and: ', newQueryBuilder.queries.$and);
            this.queries.$and = this.queries.$and.concat(newQueryBuilder.queries.$and);
        }
        else if (typeof condition === 'object') {
            Object.entries(condition).forEach(([field, value]) => {
                this.queries.$and.push({ [field]: { $eq: value } });
            });
        }
        this.lastWhere = undefined;
        return this;
    }
    sortBy(field, order) {
        this.sorter[field] = order === 'ASC' ? 1 : -1;
        return this;
    }
    paginate(config) {
        return __awaiter(this, void 0, void 0, function* () {
            this.page = Number(config.page) || 1;
            this.perPage = Number(config.perPage) || Number.MAX_SAFE_INTEGER;
            const results = yield this.get();
            const total = yield this.count();
            return {
                firstPage: 1,
                lastPage: Math.ceil(total / this.perPage),
                data: results,
                page: this.page,
                perPage: this.perPage,
                total,
            };
        });
    }
    bindRelationship(model) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.relationships && model.relationships) {
                for (const r of this.relationships) {
                    try {
                        const queryBuilder = yield model.relationships()[r]();
                        if (queryBuilder.isOne) {
                            model[r] = (yield queryBuilder.first());
                        }
                        else {
                            model[r] = (yield queryBuilder.get());
                        }
                    }
                    catch (error) {
                        console.log('error: ', error);
                        throw new Error(`Relationship "${r}" does not exists in model ${model.constructor.name}`);
                    }
                }
            }
            return model;
        });
    }
    cast(item) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!item)
                return;
            let model;
            try {
                // @ts-ignore
                model = new this.modelClass(item, this.modelClass.collectionName, true);
            }
            catch (error) {
                // @ts-ignore
                model = new this.modelClass.constructor(item, this.modelClass.collectionName, true);
            }
            model.fill(item);
            model = yield this.bindRelationship(model);
            return model;
        });
    }
    fetch() {
        return __awaiter(this, void 0, void 0, function* () {
            const query = yield this.repo.raw();
            console.log('this.queries: ', JSON.stringify(this.queries));
            let queryObject = undefined;
            if (Object.keys(this.queries.$and).length > 0) {
                queryObject = this.queries;
            }
            let cursor = query.find(queryObject);
            if (this.sorter) {
                cursor = cursor.sort(this.sorter);
            }
            if (this.page !== undefined) {
                const perPage = this.perPage || Number.MAX_SAFE_INTEGER;
                cursor = cursor.limit(perPage).skip((this.page - 1) * perPage);
            }
            if (this.isOne) {
                cursor = cursor.limit(1);
            }
            return yield cursor.toArray();
        });
    }
    get() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.fetch();
            const output = [];
            for (let item of result) {
                item = (yield this.cast(item));
                output.push(item);
            }
            return output;
        });
    }
    first() {
        return __awaiter(this, void 0, void 0, function* () {
            this.isOne = true;
            const result = (yield this.fetch())[0];
            if (!result)
                return undefined;
            return this.cast(result);
        });
    }
    count() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.repo.raw().countDocuments(this.queries, {});
        });
    }
    delete() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.repo.raw().deleteMany(this.queries, {});
        });
    }
}
exports.QueryBuilder = QueryBuilder;
//# sourceMappingURL=QueryBuilder.js.map